"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Due to the huge amount of variables involved in testing different
    rulerships, terms, and triplicities as well as testing every planet
    in each of them, this is a fairly basic set of tests using the default
    test DOB with a select few planets, and checked against Astro Gold on iOS.

"""

import copy
from datetime import datetime

from pytest import fixture

from immanuel.const import chart, dignities
from immanuel.reports import dignity
from immanuel.setup import settings
from immanuel.tools import convert, date, ephemeris


@fixture
def coords():
    # San Diego coords as used by Astro Gold
    return [convert.string_to_dec(v) for v in ('32°42\'55"', '-117°09\'23"')]

@fixture
def jd(coords):
    return date.to_jd('2000-01-01 10:00', *coords)

@fixture
def objects(jd):
    objects = (chart.SUN, chart.MOON, chart.MERCURY, chart.VENUS, chart.MARS, chart.JUPITER, chart.SATURN, chart.URANUS, chart.NEPTUNE, chart.PLUTO)
    return ephemeris.objects(objects, jd)

@fixture
def is_daytime(jd, coords):
    return ephemeris.is_daytime(jd, *coords)


def teardown_function():
    settings.include_participatory_triplicities = False
    settings.include_mutual_receptions = True


def test_ruler(objects):
    # Uranus is ruler of its sign
    assert dignity.ruler(objects[chart.SUN]) == False
    assert dignity.ruler(objects[chart.URANUS]) == True


def test_exalted(objects):
    # Nothing is exalted so we fudge the data
    sun = objects[chart.SUN].copy()
    assert dignity.exalted(sun) == False
    sun['lon'] = 0                          # Force it into Aries
    assert dignity.exalted(sun) == True


def test_triplicity_ruler(objects, is_daytime):
    # Nothing is triplicity ruler without participatory so we fudge the data
    sun = objects[chart.SUN].copy()
    assert dignity.triplicity_ruler(sun, is_daytime) == False
    sun['lon'] = 0                          # Force it into Aries
    assert dignity.triplicity_ruler(sun, is_daytime) == True

    # Saturn should be triplicity ruler with participatory included
    settings.include_participatory_triplicities = True
    assert dignity.triplicity_ruler(objects[chart.SATURN], is_daytime) == True


def test_term_ruler(objects):
    # Mercury is a term ruler
    assert dignity.term_ruler(objects[chart.SUN]) == False
    assert dignity.term_ruler(objects[chart.MERCURY]) == True


def test_face_ruler(objects):
    # Nothing is face ruler so we fudge the data
    sun = objects[chart.SUN].copy()
    assert dignity.face_ruler(sun) == False
    sun['lon'] = 10                         # Force it into 10° Aries
    assert dignity.face_ruler(sun) == True


def test_mutual_reception_ruler(objects):
    # Nothing is ruler by mutual reception so we fudge the data
    test_objects = copy.deepcopy(objects)
    sun = test_objects[chart.SUN]
    moon = test_objects[chart.MOON]
    assert dignity.mutual_reception_ruler(sun, test_objects) == False
    assert dignity.mutual_reception_ruler(moon, test_objects) == False
    sun['lon'] = 90                         # Force Sun into Cancer
    moon['lon'] = 120                       # Force Moon into Leo
    assert dignity.mutual_reception_ruler(sun, test_objects) == True
    assert dignity.mutual_reception_ruler(moon, test_objects) == True


def test_mutual_reception_exalted(objects):
    # Nothing is exalted by mutual reception so we fudge the data
    test_objects = copy.deepcopy(objects)
    sun = test_objects[chart.SUN]
    moon = test_objects[chart.MOON]
    assert dignity.mutual_reception_exalted(sun, test_objects) == False
    assert dignity.mutual_reception_exalted(moon, test_objects) == False
    sun['lon'] = 30                         # Force Sun into Taurus
    moon['lon'] = 0                         # Force Moon into Aries
    assert dignity.mutual_reception_exalted(sun, test_objects) == True
    assert dignity.mutual_reception_exalted(moon, test_objects) == True


def test_mutual_reception_triplicity_ruler(objects, is_daytime):
    # Sun & Venus are triplicity rulers by mutual reception
    assert dignity.mutual_reception_triplicity_ruler(objects[chart.SUN], objects, is_daytime) == True
    assert dignity.mutual_reception_triplicity_ruler(objects[chart.VENUS], objects, is_daytime) == True
    assert dignity.mutual_reception_triplicity_ruler(objects[chart.MOON], objects, is_daytime) == False


def test_mutual_reception_term_ruler(objects):
    # Nothing is term ruler by mutual reception so we fudge the data
    test_objects = copy.deepcopy(objects)
    mars = test_objects[chart.MARS]
    venus = test_objects[chart.VENUS]
    assert dignity.mutual_reception_term_ruler(mars, test_objects) == False
    assert dignity.mutual_reception_term_ruler(venus, test_objects) == False
    mars['lon'] = 6                         # Force Mars into 6° Aries
    venus['lon'] = 57                       # Force venus into 27° Taurus
    assert dignity.mutual_reception_term_ruler(mars, test_objects) == True
    assert dignity.mutual_reception_term_ruler(venus, test_objects) == True


def test_mutual_reception_face_ruler(objects):
    # Nothing is face ruler by mutual reception so we fudge the data
    test_objects = copy.deepcopy(objects)
    sun = test_objects[chart.SUN]
    moon = test_objects[chart.MOON]
    assert dignity.mutual_reception_face_ruler(sun, test_objects) == False
    assert dignity.mutual_reception_face_ruler(moon, test_objects) == False
    sun['lon'] = 40                         # Force Sun into 10° Taurus
    moon['lon'] = 10                        # Force Moon into 10° Aries
    assert dignity.mutual_reception_face_ruler(sun, test_objects) == True
    assert dignity.mutual_reception_face_ruler(moon, test_objects) == True


def test_in_rulership_element(objects):
    # Moon is in a Water sign, Sun is in an Earth sign
    assert dignity.in_rulership_element(objects[chart.SUN]) == False
    assert dignity.in_rulership_element(objects[chart.MOON]) == True


def test_detriment(objects):
    # Nothing is in detriment so we fudge the data
    sun = objects[chart.SUN].copy()
    assert dignity.detriment(sun) == False
    sun['lon'] = 300                        # Force it into Aquarius
    assert dignity.detriment(sun) == True


def test_fall(objects):
    # Moon is in fall in Scorpio
    assert dignity.fall(objects[chart.SUN]) == False
    assert dignity.fall(objects[chart.MOON]) == True


def test_all(objects, is_daytime):
    all = {}

    for object in objects.values():
        all[object['index']] = dignity.all(object, objects, is_daytime)

    assert all[chart.SUN][dignities.MUTUAL_RECEPTION_TRIPLICITY_RULER] == True
    assert all[chart.SUN][dignities.PEREGRINE] == False
    assert all[chart.MOON][dignities.IN_RULERSHIP_ELEMENT] == True
    assert all[chart.MOON][dignities.FALL] == True
    assert all[chart.MERCURY][dignities.TERM_RULER] == True
    assert all[chart.VENUS][dignities.MUTUAL_RECEPTION_TRIPLICITY_RULER] == True
    assert all[chart.VENUS][dignities.PEREGRINE] == False
    assert all[chart.MARS][dignities.PEREGRINE] == True
    assert all[chart.JUPITER][dignities.IN_RULERSHIP_ELEMENT] == True
    assert all[chart.SATURN][dignities.IN_RULERSHIP_ELEMENT] == True
    assert all[chart.URANUS][dignities.RULER] == True
    assert all[chart.URANUS][dignities.IN_RULERSHIP_ELEMENT] == True
    assert all[chart.NEPTUNE][dignities.FALL] == True
    assert all[chart.NEPTUNE][dignities.PEREGRINE] == True
    assert all[chart.PLUTO][dignities.PEREGRINE] == True

    settings.include_mutual_receptions = False

    for object in objects.values():
        all[object['index']] = dignity.all(object, objects, is_daytime)

    assert all[chart.SUN][dignities.MUTUAL_RECEPTION_TRIPLICITY_RULER] == True
    assert all[chart.SUN][dignities.PEREGRINE] == True
    assert all[chart.MOON][dignities.IN_RULERSHIP_ELEMENT] == True
    assert all[chart.MOON][dignities.FALL] == True
    assert all[chart.MERCURY][dignities.TERM_RULER] == True
    assert all[chart.VENUS][dignities.MUTUAL_RECEPTION_TRIPLICITY_RULER] == True
    assert all[chart.VENUS][dignities.PEREGRINE] == True
    assert all[chart.MARS][dignities.PEREGRINE] == True
    assert all[chart.JUPITER][dignities.IN_RULERSHIP_ELEMENT] == True
    assert all[chart.SATURN][dignities.IN_RULERSHIP_ELEMENT] == True
    assert all[chart.URANUS][dignities.RULER] == True
    assert all[chart.URANUS][dignities.IN_RULERSHIP_ELEMENT] == True
    assert all[chart.NEPTUNE][dignities.FALL] == True
    assert all[chart.NEPTUNE][dignities.PEREGRINE] == True
    assert all[chart.PLUTO][dignities.PEREGRINE] == True


def test_score(objects, is_daytime):
    # Astro Gold only calculates scores for the main 7 planets
    scores = {}

    for object in objects.values():
        scores[object['index']] = dignity.score(dignity.all(object, objects, is_daytime))

    assert scores[chart.SUN] == 3
    assert scores[chart.MOON] == -4
    assert scores[chart.MERCURY] == 2
    assert scores[chart.VENUS] == 3
    assert scores[chart.MARS] == -5
    assert scores[chart.JUPITER] == 0
    assert scores[chart.SATURN] == 0

    settings.include_mutual_receptions = False

    for object in objects.values():
        scores[object['index']] = dignity.score(dignity.all(object, objects, is_daytime))

    assert scores[chart.SUN] == -2
    assert scores[chart.MOON] == -4
    assert scores[chart.MERCURY] == 2
    assert scores[chart.VENUS] == -2
    assert scores[chart.MARS] == -5
    assert scores[chart.JUPITER] == 0
    assert scores[chart.SATURN] == 0



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    The calculate module's figures are tested against output
    from astro.com and https://stardate.org/nightsky/moon
    for the moon_phase() function.

"""

from datetime import datetime

from pytest import fixture

from immanuel.const import calc, chart
from immanuel.tools import calculate, convert, date, ephemeris, position


@fixture
def coords():
    # San Diego coords as used by astro.com
    return [convert.string_to_dec(v) for v in ('32n43', '117w09')]

@fixture
def day_jd(coords):
    return date.to_jd('2000-01-01 10:00', *coords)

@fixture
def night_jd(coords):
    return date.to_jd('2000-01-01 00:00', *coords)


def test_moon_phase(day_jd):
    # Courtesy of https://stardate.org/nightsky/moon
    sun = ephemeris.planet(chart.SUN, day_jd)
    moon = ephemeris.planet(chart.MOON, day_jd)
    assert calculate.moon_phase(sun, moon) == calc.THIRD_QUARTER             # third quarter = waning crescent


def test_is_daytime(day_jd, night_jd, coords):
    sun, asc = ephemeris.objects((chart.SUN, chart.ASC), day_jd, *coords, chart.PLACIDUS).values()
    assert calculate.is_daytime(sun, asc) == True
    sun, asc = ephemeris.objects((chart.SUN, chart.ASC), night_jd, *coords, chart.PLACIDUS).values()
    assert calculate.is_daytime(sun, asc) == False


def test_pars_fortuna_day_formula(day_jd, coords):
    sun, moon, asc = ephemeris.objects((chart.SUN, chart.MOON, chart.ASC), day_jd, *coords, chart.PLACIDUS).values()
    pof = calculate.pars_fortuna_longitude(sun, moon, asc, calc.DAY_FORMULA)
    sign = position.sign(pof)
    lon = position.sign_longitude(pof)
    assert sign == chart.CAPRICORN
    assert convert.dec_to_string(lon) == '11°18\'41"'


def test_pars_fortuna_night_formula(night_jd, coords):
    sun, moon, asc = ephemeris.objects((chart.SUN, chart.MOON, chart.ASC), night_jd, *coords, chart.PLACIDUS).values()
    pof = calculate.pars_fortuna_longitude(sun, moon, asc, calc.NIGHT_FORMULA)
    sign = position.sign(pof)
    lon = position.sign_longitude(pof)
    assert sign == chart.SAGITTARIUS
    assert convert.dec_to_string(lon) == '10°04\'30"'


def test_sidereal_time(day_jd, coords):
    armc = ephemeris.angle(chart.ARMC, day_jd, *coords, chart.PLACIDUS)
    sidereal_time = calculate.sidereal_time(armc)
    assert convert.dec_to_string(sidereal_time, convert.FORMAT_TIME) == '16:54:13'


def test_object_movement(day_jd, coords):
    sun, moon, saturn, true_north_node, pars_fortuna = ephemeris.objects((chart.SUN, chart.MOON, chart.SATURN, chart.TRUE_NORTH_NODE, chart.PARS_FORTUNA), day_jd, *coords, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA).values()
    assert calculate.object_movement(sun) == calc.DIRECT
    assert calculate.object_movement(moon) == calc.DIRECT
    assert calculate.object_movement(saturn) == calc.RETROGRADE
    assert calculate.object_movement(true_north_node) == calc.RETROGRADE
    assert calculate.object_movement(pars_fortuna) == calc.STATIONARY


def test_is_out_of_bounds(day_jd, coords):
    sun, mercury = ephemeris.objects((chart.SUN, chart.MERCURY), day_jd, *coords, chart.PLACIDUS).values()
    assert calculate.is_out_of_bounds(sun, day_jd) == False
    assert calculate.is_out_of_bounds(mercury, day_jd) == True


def test_solar_year_length():
    """ This one is difficult to test in isolatioon since it's only used for
    secondary progressions. For now we'll leave it to the forecast module
    tests to check the correct progressed Julian dates. """
    pass



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Test various birth charts with known chart shapes.
    Celebrity natal chart data & chart shapes courtesy of
    https://horoscopes.astro-seek.com

"""

from datetime import datetime

from pytest import fixture

from immanuel.const import calc, chart
from immanuel.reports import pattern
from immanuel.tools import convert, date, ephemeris


@fixture
def object_indices():
    return (
        chart.SUN,
        chart.MOON,
        chart.MERCURY,
        chart.VENUS,
        chart.MARS,
        chart.JUPITER,
        chart.SATURN,
        chart.URANUS,
        chart.NEPTUNE,
        chart.PLUTO,
    )

@fixture
def birth_data():
    return {
        # Harrison Ford
        calc.BUNDLE: {
            'latitude': '41n51',
            'longitude': '87w39',
            'dob': '1942-07-13 11:41:00',
        },
        # Clint Eastwood
        calc.BUCKET: {
            'latitude': '37n47',
            'longitude': '122w25',
            'dob': '1930-05-31 17:35:00',
        },
        # Alfred Hitchcock
        calc.BOWL: {
            'latitude': '51n30',
            'longitude': '0w10',
            'dob': '1899-08-13 20:00:00',
        },
        # Isaac Newton
        calc.LOCOMOTIVE: {
            'latitude': '52n49',
            'longitude': '0w38',
            'dob': '1643-01-04 01:38:00',
        },
        # William Blake
        calc.SEESAW: {
            'latitude': '51n30',
            'longitude': '0w08',
            'dob': '1757-11-28 19:45:00',
        },
        # Carl Jung
        calc.SPLASH: {
            'latitude': '46n36',
            'longitude': '9e19',
            'dob': '1875-07-26 19:29:00',
        },
        # Random DOB to get a non-shape
        calc.SPLAY: {
            'latitude': '32n43',
            'longitude': '117w09',
            'dob': '1902-01-01 10:00:00',
        }
    }


def test_chart_shape(object_indices, birth_data):
    for chart_shape, data in birth_data.items():
        lat, lon = (convert.string_to_dec(v) for v in (data['latitude'], data['longitude']))
        dob_dt = date.localize(datetime.fromisoformat(data['dob']), lat, lon)
        jd = date.to_jd(dob_dt)
        objects = ephemeris.objects(object_indices, jd, lat, lon, chart.PLACIDUS)
        assert pattern.chart_shape(objects) == chart_shape



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Test a natal chart's object distribution against astro.com's reports.

"""

from datetime import datetime

from pytest import fixture

from immanuel.const import chart
from immanuel.reports import weighting
from immanuel.tools import convert, date, ephemeris


@fixture
def coords():
    # San Diego coords as used by astro.com
    return [convert.string_to_dec(v) for v in ('32n43', '117w09')]

@fixture
def jd(coords):
    return date.to_jd('2000-01-01 10:00', *coords)

@fixture
def objects(coords, jd):
    object_indices = (
        chart.ASC,
        chart.MC,
        chart.SUN,
        chart.MOON,
        chart.MERCURY,
        chart.VENUS,
        chart.MARS,
        chart.JUPITER,
        chart.SATURN,
        chart.URANUS,
        chart.NEPTUNE,
        chart.PLUTO,
        chart.TRUE_NORTH_NODE,
        chart.CHIRON,
    )
    return ephemeris.objects(object_indices, jd, *coords, chart.PLACIDUS)

@fixture
def houses(coords, jd):
    return ephemeris.houses(jd, *coords, chart.PLACIDUS)


def test_elements(objects):
    elements = weighting.elements(objects)
    assert elements[chart.FIRE] == [chart.MC, chart.VENUS, chart.JUPITER, chart.PLUTO, chart.TRUE_NORTH_NODE, chart.CHIRON]
    assert elements[chart.EARTH] == [chart.SUN, chart.MERCURY, chart.SATURN]
    assert elements[chart.AIR] == [chart.MARS, chart.URANUS, chart.NEPTUNE]
    assert elements[chart.WATER] == [chart.ASC, chart.MOON]


def test_modalities(objects):
    modalities = weighting.modalities(objects)
    assert modalities[chart.CARDINAL] == [chart.SUN, chart.MERCURY, chart.JUPITER]
    assert modalities[chart.FIXED] == [chart.MOON, chart.MARS, chart.SATURN, chart.URANUS, chart.NEPTUNE, chart.TRUE_NORTH_NODE]
    assert modalities[chart.MUTABLE] == [chart.ASC, chart.MC, chart.VENUS, chart.PLUTO, chart.CHIRON]


def test_quadrants(objects, houses):
    quadrants = weighting.quadrants(objects, houses)
    assert quadrants[1] == [chart.ASC, chart.JUPITER, chart.SATURN]
    assert quadrants[2] == [chart.TRUE_NORTH_NODE]
    assert quadrants[3] == [chart.MOON, chart.VENUS, chart.PLUTO, chart.CHIRON]
    assert quadrants[4] == [chart.MC, chart.SUN, chart.MERCURY, chart.MARS, chart.URANUS, chart.NEPTUNE]



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    The actual user-facing chart classes are contained in this module. Each
    chart class is easily serializable using the ToJSON class. Each chart type
    is instantiated by passing an instance of Subject, apart from Transits.
    This assumes the current moment and optionally takes a pair of coordinates
    for house calculations, although these will default to those specifed in
    the settings if they are not required.

    Instead of a dedicated synastry chart, the optional aspects_to parameter
    in each chart type's constructor takes another Chart instance and forms its
    aspects to the planets in that chart rather than its own.

"""

from datetime import datetime
from zoneinfo import ZoneInfo

from immanuel.classes import wrap
from immanuel.const import chart, names
from immanuel.reports import aspect, pattern, weighting
from immanuel.setup import settings
from immanuel.tools import calculate, convert, date, ephemeris, forecast, midpoint


class Subject:
    """ Simple class to model a chart subject - essentially just
    a time and place. """
    def __init__(self, date_time: datetime | str, latitude: float | list | tuple | str, longitude: float | list | tuple | str, time_is_dst: bool = None) -> None:
        self.latitude, self.longitude = (convert.to_dec(v) for v in (latitude, longitude))
        self.time_is_dst = time_is_dst
        self.date_time = date.to_datetime(
                dt=date_time,
                lat=self.latitude,
                lon=self.longitude,
                is_dst=time_is_dst
            )
        self.julian_date = date.to_jd(self.date_time)


class Chart():
    """ Base chart class. This acts as an abstract class for the actual chart
    classes to inherit from. """
    def __init__(self, type: int, aspects_to: 'Chart' = None) -> None:
        self.type = names.CHART_TYPES[type]
        self._type = type
        self._aspects_to = aspects_to
        self.generate()
        self.wrap()

    def generate(self) -> None:
        """ Generating the raw data is each descendant class's responsibility,
        but placeholders for properties common to all charts are set here. """
        self._native: Subject = None
        self._obliquity: float = None
        self._diurnal: bool = None
        self._moon_phase: int = None
        self._objects: dict = {}
        self._houses: dict = {}

    def wrap(self) -> None:
        """ Loop through the required data and wrap each one with a custom
        function. """
        for index in settings.chart_data[self._type]:
            method = f'set_wrapped_{index}'
            if hasattr(self, method):
                getattr(self, method)()

    # Base class provides wrappers for properties common to all classes.
    def set_wrapped_native(self) -> None:
        self.native = wrap.Subject(self._native)

    def set_wrapped_house_system(self) -> None:
        self.house_system = names.HOUSE_SYSTEMS[settings.house_system]

    def set_wrapped_shape(self) -> None:
        self.shape = names.CHART_SHAPES[pattern.chart_shape(self._objects)]

    def set_wrapped_diurnal(self) -> None:
        self.diurnal = self._diurnal

    def set_wrapped_moon_phase(self) -> None:
        self.moon_phase = wrap.MoonPhase(self._moon_phase)

    def set_wrapped_objects(self) -> None:
        self.objects = {}
        for index, object in self._objects.items():
            if 'jd' in object:
                object['date_time'] = date.to_datetime(
                        dt=object['jd'],
                        lat=self._native.latitude,
                        lon=self._native.longitude,
                    )
            self.objects[index] = wrap.Object(
                    object=object,
                    objects=self._objects,
                    houses=self._houses,
                    is_daytime=self._diurnal,
                    obliquity=self._obliquity,
                )

    def set_wrapped_houses(self) -> None:
        self.houses = {index: wrap.Object(object=house) for index, house in self._houses.items()}

    def set_wrapped_aspects(self) -> None:
        aspects = aspect.all(self._objects) if self._aspects_to is None else aspect.synastry(self._objects, self._aspects_to._objects)
        self.aspects = {index: {object_index: wrap.Aspect(aspect=object_aspect, objects=self._objects) for object_index, object_aspect in aspect_list.items()} for index, aspect_list in aspects.items()}

    def set_wrapped_weightings(self) -> None:
        self.weightings = {
            'elements': wrap.Elements(weighting.elements(self._objects)),
            'modalities': wrap.Modalities(weighting.modalities(self._objects)),
            'quadrants': wrap.Quadrants(weighting.quadrants(self._objects, self._houses)),
        }


class Natal(Chart):
    """ Standard natal chart generates data straight from the passed
    native information. """
    def __init__(self, native: Subject, aspects_to: Chart = None) -> None:
        self._native = native
        super().__init__(chart.NATAL, aspects_to)

    def generate(self) -> None:
        self._obliquity = ephemeris.obliquity(self._native.julian_date)

        sun = ephemeris.planet(chart.SUN, self._native.julian_date)
        moon = ephemeris.planet(chart.MOON, self._native.julian_date)
        asc = ephemeris.angle(
                index=chart.ASC,
                jd=self._native.julian_date,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
            )

        self._diurnal = calculate.is_daytime(sun, asc)
        self._moon_phase = calculate.moon_phase(sun, moon)
        self._objects = ephemeris.objects(
                object_list=settings.objects,
                jd=self._native.julian_date,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
                pars_fortuna_formula=settings.pars_fortuna_formula,
            )
        self._houses = ephemeris.houses(
                jd=self._native.julian_date,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
            )


class SolarReturn(Chart):
    """ Solar return chart for the given year. """
    def __init__(self, native: Subject, year: int, aspects_to: Chart = None) -> None:
        self._native = native
        self._solar_return_year = year
        super().__init__(chart.SOLAR_RETURN, aspects_to)

    def generate(self) -> None:
        self._solar_return_jd = forecast.solar_return(self._native.julian_date, self._solar_return_year)
        self._obliquity = ephemeris.obliquity(self._solar_return_jd)
        self._solar_return_armc = ephemeris.angle(
                index=chart.ARMC,
                jd=self._solar_return_jd,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
            )

        sun = ephemeris.planet(chart.SUN, self._solar_return_jd)
        moon = ephemeris.planet(chart.MOON, self._solar_return_jd)
        asc = ephemeris.angle(
                index=chart.ASC,
                jd=self._solar_return_jd,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
            )

        self._diurnal = calculate.is_daytime(sun, asc)
        self._moon_phase = calculate.moon_phase(sun, moon)
        self._objects = ephemeris.objects(
                object_list=settings.objects,
                jd=self._solar_return_jd,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
                pars_fortuna_formula=settings.pars_fortuna_formula,
            )
        self._houses = ephemeris.houses(
                jd=self._solar_return_jd,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
            )

    def set_wrapped_solar_return_year(self) -> None:
        self.solar_return_year = self._solar_return_year

    def set_wrapped_solar_return_date_time(self) -> None:
        self.solar_return_date_time = wrap.DateTime(
                dt=self._solar_return_jd,
                armc=self._solar_return_armc,
                latitude=self._native.latitude,
                longitude=self._native.longitude,
            )


class Progressed(Chart):
    """ Secondary progression chart uses the MC progression method from
    settings. """
    def __init__(self, native: Subject, date_time: datetime | str, aspects_to: Chart = None) -> None:
        self._native = native
        self._date_time = date_time
        super().__init__(chart.PROGRESSED, aspects_to)

    def generate(self) -> None:
        self._progression_date_time = date.to_datetime(
                dt=self._date_time,
                lat=self._native.latitude,
                lon=self._native.longitude,
            )
        progression_jd = date.to_jd(self._progression_date_time)
        progression_armc = ephemeris.angle(
                index=chart.ARMC,
                jd=progression_jd,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
            )
        self._progression_armc_longitude = progression_armc['lon']

        self._progressed_jd, self._progressed_armc_longitude = forecast.progression(
                jd=self._native.julian_date,
                lat=self._native.latitude,
                lon=self._native.longitude,
                pjd=progression_jd,
                house_system=settings.house_system,
                method=settings.mc_progression_method,
            )
        self._obliquity = ephemeris.obliquity(self._progressed_jd)

        sun = ephemeris.planet(chart.SUN, self._progressed_jd)
        moon = ephemeris.planet(chart.MOON, self._progressed_jd)
        asc = ephemeris.armc_angle(
            index=chart.ASC,
                armc=self._progressed_armc_longitude,
                lat=self._native.latitude,
                obliquity=self._obliquity,
                house_system=settings.house_system,
            )

        self._diurnal = calculate.is_daytime(sun, asc)
        self._moon_phase = calculate.moon_phase(sun, moon)
        self._objects = ephemeris.armc_objects(
                object_list=settings.objects,
                jd=self._progressed_jd,
                armc=self._progressed_armc_longitude,
                lat=self._native.latitude,
                lon=self._native.longitude,
                obliquity=self._obliquity,
                house_system=settings.house_system,
                pars_fortuna_formula=settings.pars_fortuna_formula,
            )
        self._houses = ephemeris.armc_houses(
                armc=self._progressed_armc_longitude,
                lat=self._native.latitude,
                obliquity=self._obliquity,
                house_system=settings.house_system,
            )

    def set_wrapped_progression_date_time(self) -> None:
        self.progression_date_time = wrap.DateTime(
                dt=self._progression_date_time,
                armc=self._progression_armc_longitude,
            )

    def set_wrapped_progressed_date_time(self) -> None:
        self.progressed_date_time = wrap.DateTime(
                dt=self._progressed_jd,
                armc=self._progressed_armc_longitude,
                latitude=self._native.latitude,
                longitude=self._native.longitude,
            )

    def set_wrapped_progression_method(self) -> None:
        self.progression_method = names.PROGRESSION_METHODS[settings.mc_progression_method]


class Composite(Chart):
    """ Generates a midpoint chart based on the two passed sets of data. """
    def __init__(self, native: Subject, partner: Subject, aspects_to: Chart = None) -> None:
        self._native = native
        self._partner = partner
        super().__init__(chart.COMPOSITE, aspects_to)

    def generate(self) -> None:
        self._obliquity = midpoint.obliquity(self._native.julian_date, self._partner.julian_date)

        native_objects = ephemeris.objects(
                object_list=settings.objects,
                jd=self._native.julian_date,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
                pars_fortuna_formula=settings.pars_fortuna_formula,
            )
        partner_objects = ephemeris.objects(
                object_list=settings.objects,
                jd=self._partner.julian_date,
                lat=self._partner.latitude,
                lon=self._partner.longitude,
                house_system=settings.house_system,
                pars_fortuna_formula=settings.pars_fortuna_formula,
            )
        self._objects = midpoint.all(
                objects1=native_objects,
                objects2=partner_objects,
                obliquity=self._obliquity,
            )

        native_houses = ephemeris.houses(
                jd=self._native.julian_date,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
            )
        partner_houses = ephemeris.houses(
                jd=self._partner.julian_date,
                lat=self._partner.latitude,
                lon=self._partner.longitude,
                house_system=settings.house_system,
            )
        self._houses = midpoint.all(
                objects1=native_houses,
                objects2=partner_houses,
                obliquity=self._obliquity,
            )

        native_sun = ephemeris.planet(chart.SUN, self._native.julian_date)
        native_moon = ephemeris.planet(chart.MOON, self._native.julian_date)

        partner_sun = ephemeris.planet(chart.SUN, self._partner.julian_date)
        partner_moon = ephemeris.planet(chart.MOON, self._partner.julian_date)

        sun = midpoint.composite(native_sun, partner_sun, self._obliquity)
        moon = midpoint.composite(native_moon, partner_moon, self._obliquity)
        asc = self._houses[chart.HOUSE1]

        self._diurnal = calculate.is_daytime(sun, asc)
        self._moon_phase = calculate.moon_phase(sun, moon)

    def set_wrapped_partner(self):
        self.partner = wrap.Subject(self._partner)


class Transits(Chart):
    """ Chart of the moment for the given coordinates. Structurally identical
    to the natal chart. Coordinates default to those specified in settings. """
    def __init__(self, latitude: float | list | tuple | str = settings.default_latitude, longitude: float | list | tuple | str = settings.default_longitude, aspects_to: Chart = None) -> None:
        lat, lon = (convert.to_dec(v) for v in (latitude, longitude))
        timezone = date.timezone(lat, lon)
        date_time = datetime.now(tz=ZoneInfo(timezone))
        self._native = Subject(date_time, lat, lon)
        super().__init__(chart.TRANSITS, aspects_to)

    def generate(self) -> None:
        self._obliquity = ephemeris.obliquity(self._native.julian_date)

        sun = ephemeris.planet(chart.SUN, self._native.julian_date)
        moon = ephemeris.planet(chart.MOON, self._native.julian_date)
        asc = ephemeris.angle(
                index=chart.ASC,
                jd=self._native.julian_date,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
            )

        self._diurnal = calculate.is_daytime(sun, asc)
        self._moon_phase = calculate.moon_phase(sun, moon)
        self._objects = ephemeris.objects(
                object_list=settings.objects,
                jd=self._native.julian_date,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
                pars_fortuna_formula=settings.pars_fortuna_formula,
            )
        self._houses = ephemeris.houses(
                jd=self._native.julian_date,
                lat=self._native.latitude,
                lon=self._native.longitude,
                house_system=settings.house_system,
            )



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    This module calculates solar returns and secondary progressions.

"""

import swisseph as swe

from immanuel.const import calc, chart
from immanuel.tools import calculate, date, ephemeris


JD = 0
ARMC = 1


def solar_return(jd: float, year: int) -> float:
    """ Returns the Julian date of the given year's solar return. """
    dt = date.to_datetime(jd)
    year_diff = year - dt.year
    sr_jd = jd + year_diff * calc.YEAR_DAYS
    natal_sun = ephemeris.planet(chart.SUN, jd)

    while True:
        sr_sun = ephemeris.planet(chart.SUN, sr_jd)
        distance = swe.difdeg2n(natal_sun['lon'], sr_sun['lon'])
        if abs(distance) <= calc.MAX_ERROR:
            break
        sr_jd += distance / sr_sun['speed']

    return sr_jd


def progression(jd: float, lat: float, lon: float, pjd: float, house_system: int, method: int) -> tuple:
    """ Returns the progressed Julian date and MC right ascension. """
    year_days = calculate.solar_year_length(jd)
    years = (pjd - jd) / year_days
    progressed_jd = jd + years

    match method:
        case calc.DAILY_HOUSES:
            progressed_armc_lon = ephemeris.angle(chart.ARMC, progressed_jd, lat, lon, house_system)['lon']
        case calc.NAIBOD:
            natal_armc = ephemeris.angle(chart.ARMC, jd, lat, lon, house_system)['lon']
            progressed_armc_lon = swe.degnorm(natal_armc + years * calc.MEAN_MOTIONS[chart.SUN])
        case calc.SOLAR_ARC:
            natal_mc = ephemeris.angle(chart.MC, jd, lat, lon, house_system)
            natal_sun = ephemeris.planet(chart.SUN, jd)
            progressed_sun = ephemeris.planet(chart.SUN, progressed_jd)
            distance = swe.difdeg2n(progressed_sun['lon'], natal_sun['lon'])
            obliquity = ephemeris.obliquity(progressed_jd)
            progressed_armc_lon = swe.cotrans((natal_mc['lon'] + distance, 0, 1), -obliquity)[0]

    return progressed_jd, progressed_armc_lon



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    DMS / decimal tests ran against figures from:
    https://www.rapidtables.com/convert/number/degrees-minutes-seconds-to-degrees.html

"""

from immanuel.tools import convert


""" dms_to_dec() """
def test_dms_to_dec():
    assert convert.dms_to_dec(('+', 12, 30, 45)) == 12.5125
    assert convert.dms_to_dec(('+', 12, 30.75)) == 12.5125
    assert convert.dms_to_dec(('-', 12, 30, 45)) == -12.5125
    assert convert.dms_to_dec(('-', 12, 30.75)) == -12.5125


""" dec_to_dms() """
def test_dec_to_dms_round_degree():
    assert convert.dec_to_dms(12.5125, convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.dec_to_dms(12.5125, convert.ROUND_DEGREE, False) == ('+', 13)
    assert convert.dec_to_dms(-12.5125, convert.ROUND_DEGREE, True) == ('-', 13, 0, 0)
    assert convert.dec_to_dms(-12.5125, convert.ROUND_DEGREE, False) == ('-', 13)


def test_dec_to_dms_round_minute():
    assert convert.dec_to_dms(12.5125, convert.ROUND_MINUTE, True) == ('+', 12, 31, 0)
    assert convert.dec_to_dms(12.5125, convert.ROUND_MINUTE, False) == ('+', 12, 31)
    assert convert.dec_to_dms(-12.5125, convert.ROUND_MINUTE, True) == ('-', 12, 31, 0)
    assert convert.dec_to_dms(-12.5125, convert.ROUND_MINUTE, False) == ('-', 12, 31)


def test_dec_to_dms_round_second():
    assert convert.dec_to_dms(12.5125, convert.ROUND_SECOND, True) == ('+', 12, 30, 45)
    assert convert.dec_to_dms(12.5125, convert.ROUND_SECOND, False) == ('+', 12, 30, 45)
    assert convert.dec_to_dms(-12.5125, convert.ROUND_SECOND, True) == ('-', 12, 30, 45)
    assert convert.dec_to_dms(-12.5125, convert.ROUND_SECOND, False) == ('-', 12, 30, 45)


""" dms_to_string() """
def test_dms_to_string_time_round_degree():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_TIME, convert.ROUND_DEGREE) == '13:00:00'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_TIME, convert.ROUND_DEGREE) == '13:00:00'


def test_dms_to_string_time_round_minute():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_TIME, convert.ROUND_MINUTE) == '12:31:00'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_TIME, convert.ROUND_MINUTE) == '12:31:00'


def test_dms_to_string_time_round_second():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_TIME, convert.ROUND_SECOND) == '12:30:45'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_TIME, convert.ROUND_SECOND) == '12:30:45'


def test_dms_to_string_time_offset_round_degree():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '+13:00:00'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '+13:00:00'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '-13:00:00'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '-13:00:00'


def test_dms_to_string_time_offset_round_minute():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_TIME_OFFSET, convert.ROUND_MINUTE) == '+12:31:00'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_TIME_OFFSET, convert.ROUND_MINUTE) == '+12:31:00'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_TIME_OFFSET, convert.ROUND_MINUTE) == '-12:31:00'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_TIME_OFFSET, convert.ROUND_MINUTE) == '-12:31:00'


def test_dms_to_string_time_offset_round_second():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_TIME_OFFSET, convert.ROUND_SECOND) == '+12:30:45'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_TIME_OFFSET, convert.ROUND_SECOND) == '+12:30:45'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_TIME_OFFSET, convert.ROUND_SECOND) == '-12:30:45'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_TIME_OFFSET, convert.ROUND_SECOND) == '-12:30:45'


def test_dms_to_string_dms_round_degree():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_DEGREE) == '13°'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_DEGREE) == '13°'
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '13°00\'00"'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '13°00\'00"'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_DEGREE) == '-13°'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_DEGREE) == '-13°'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '-13°00\'00"'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '-13°00\'00"'


def test_dms_to_string_dms_round_minute():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_MINUTE) == '12°31\''
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_MINUTE) == '12°31\''
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_MINUTE, True) == '12°31\'00"'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_MINUTE, True) == '12°31\'00"'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_MINUTE) == '-12°31\''
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_MINUTE) == '-12°31\''
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_MINUTE, True) == '-12°31\'00"'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_MINUTE, True) == '-12°31\'00"'


def test_dms_to_string_dms_round_second():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_SECOND) == '12°30\'45"'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_SECOND) == '12°30\'45"'
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_SECOND, True) == '12°30\'45"'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_SECOND, True) == '12°30\'45"'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_SECOND) == '-12°30\'45"'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_SECOND) == '-12°30\'45"'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_DMS, convert.ROUND_SECOND, True) == '-12°30\'45"'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_DMS, convert.ROUND_SECOND, True) == '-12°30\'45"'


def test_dms_to_string_lat_round_degree():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_LAT, convert.ROUND_DEGREE) == '13N0.0'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_LAT, convert.ROUND_DEGREE) == '13N0.0'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_LAT, convert.ROUND_DEGREE) == '13S0.0'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_LAT, convert.ROUND_DEGREE) == '13S0.0'


def test_dms_to_string_lat_round_minute():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_LAT, convert.ROUND_MINUTE) == '12N31.0'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_LAT, convert.ROUND_MINUTE) == '12N31.0'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_LAT, convert.ROUND_MINUTE) == '12S31.0'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_LAT, convert.ROUND_MINUTE) == '12S31.0'


def test_dms_to_string_lat_round_minute():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_LAT, convert.ROUND_SECOND) == '12N30.75'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_LAT, convert.ROUND_SECOND) == '12N30.75'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_LAT, convert.ROUND_SECOND) == '12S30.75'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_LAT, convert.ROUND_SECOND) == '12S30.75'


def test_dms_to_string_lon_round_degree():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_LON, convert.ROUND_DEGREE) == '13E0.0'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_LON, convert.ROUND_DEGREE) == '13E0.0'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_LON, convert.ROUND_DEGREE) == '13W0.0'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_LON, convert.ROUND_DEGREE) == '13W0.0'


def test_dms_to_string_lon_round_minute():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_LON, convert.ROUND_MINUTE) == '12E31.0'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_LON, convert.ROUND_MINUTE) == '12E31.0'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_LON, convert.ROUND_MINUTE) == '12W31.0'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_LON, convert.ROUND_MINUTE) == '12W31.0'


def test_dms_to_string_lon_round_minute():
    assert convert.dms_to_string(('+', 12, 30, 45), convert.FORMAT_LON, convert.ROUND_SECOND) == '12E30.75'
    assert convert.dms_to_string(('+', 12, 30.75), convert.FORMAT_LON, convert.ROUND_SECOND) == '12E30.75'
    assert convert.dms_to_string(('-', 12, 30, 45), convert.FORMAT_LON, convert.ROUND_SECOND) == '12W30.75'
    assert convert.dms_to_string(('-', 12, 30.75), convert.FORMAT_LON, convert.ROUND_SECOND) == '12W30.75'


""" string_to_dms() """
def test_string_to_dms_time_rounded_degree():
    assert convert.string_to_dms('12:30:45', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('12:30', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('12:30:45', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.string_to_dms('12:30', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)


def test_string_to_dms_time_rounded_minute():
    assert convert.string_to_dms('12:30:45', convert.ROUND_MINUTE) == ('+', 12, 31)
    assert convert.string_to_dms('12:30', convert.ROUND_MINUTE) == ('+', 12, 30)
    assert convert.string_to_dms('12:30:45', convert.ROUND_MINUTE, True) == ('+', 12, 31, 0)
    assert convert.string_to_dms('12:30', convert.ROUND_MINUTE, True) == ('+', 12, 30, 0)


def test_string_to_dms_time_rounded_second():
    assert convert.string_to_dms('12:30:45', convert.ROUND_SECOND) == ('+', 12, 30, 45)
    assert convert.string_to_dms('12:30:45', convert.ROUND_SECOND, True) == ('+', 12, 30, 45)


def test_string_to_dms_time_offset_rounded_degree():
    assert convert.string_to_dms('+12:30:45', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('+12:30:45', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.string_to_dms('+12:30', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('+12:30', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.string_to_dms('-12:30:45', convert.ROUND_DEGREE) == ('-', 13)
    assert convert.string_to_dms('-12:30:45', convert.ROUND_DEGREE, True) == ('-', 13, 0, 0)
    assert convert.string_to_dms('-12:30', convert.ROUND_DEGREE) == ('-', 13)
    assert convert.string_to_dms('-12:30', convert.ROUND_DEGREE, True) == ('-', 13, 0, 0)


def test_string_to_dms_time_offset_rounded_minute():
    assert convert.string_to_dms('+12:30:45', convert.ROUND_MINUTE) == ('+', 12, 31)
    assert convert.string_to_dms('+12:30:45', convert.ROUND_MINUTE, True) == ('+', 12, 31, 0)
    assert convert.string_to_dms('+12:30', convert.ROUND_MINUTE) == ('+', 12, 30)
    assert convert.string_to_dms('+12:30', convert.ROUND_MINUTE, True) == ('+', 12, 30, 0)
    assert convert.string_to_dms('-12:30:45', convert.ROUND_MINUTE) == ('-', 12, 31)
    assert convert.string_to_dms('-12:30:45', convert.ROUND_MINUTE, True) == ('-', 12, 31, 0)
    assert convert.string_to_dms('-12:30', convert.ROUND_MINUTE) == ('-', 12, 30)
    assert convert.string_to_dms('-12:30', convert.ROUND_MINUTE, True) == ('-', 12, 30, 0)


def test_string_to_dms_time_offset_rounded_second():
    assert convert.string_to_dms('+12:30:45', convert.ROUND_SECOND) == ('+', 12, 30, 45)
    assert convert.string_to_dms('+12:30:45', convert.ROUND_SECOND, True) == ('+', 12, 30, 45)
    assert convert.string_to_dms('-12:30:45', convert.ROUND_SECOND) == ('-', 12, 30, 45)
    assert convert.string_to_dms('-12:30:45', convert.ROUND_SECOND, True) == ('-', 12, 30, 45)


def test_string_to_dms_dms_rounded_degree():
    assert convert.string_to_dms('12°30\'45"', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('12°30\'45"', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.string_to_dms('12°30\'', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('12°30\'', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.string_to_dms('-12°30\'45"', convert.ROUND_DEGREE) == ('-', 13)
    assert convert.string_to_dms('-12°30\'45"', convert.ROUND_DEGREE, True) == ('-', 13, 0, 0)
    assert convert.string_to_dms('-12°30\'', convert.ROUND_DEGREE) == ('-', 13)
    assert convert.string_to_dms('-12°30\'', convert.ROUND_DEGREE, True) == ('-', 13, 0, 0)


def test_string_to_dms_dms_rounded_minute():
    assert convert.string_to_dms('12°30\'45"', convert.ROUND_MINUTE) == ('+', 12, 31)
    assert convert.string_to_dms('12°30\'45"', convert.ROUND_MINUTE, True) == ('+', 12, 31, 0)
    assert convert.string_to_dms('12°30\'', convert.ROUND_MINUTE) == ('+', 12, 30)
    assert convert.string_to_dms('12°30\'', convert.ROUND_MINUTE, True) == ('+', 12, 30, 0)
    assert convert.string_to_dms('-12°30\'45"', convert.ROUND_MINUTE) == ('-', 12, 31)
    assert convert.string_to_dms('-12°30\'45"', convert.ROUND_MINUTE, True) == ('-', 12, 31, 0)
    assert convert.string_to_dms('-12°30\'', convert.ROUND_MINUTE) == ('-', 12, 30)
    assert convert.string_to_dms('-12°30\'', convert.ROUND_MINUTE, True) == ('-', 12, 30, 0)


def test_string_to_dms_dms_rounded_second():
    assert convert.string_to_dms('+12°30\'45"', convert.ROUND_SECOND) == ('+', 12, 30, 45)
    assert convert.string_to_dms('+12°30\'45"', convert.ROUND_SECOND, True) == ('+', 12, 30, 45)
    assert convert.string_to_dms('-12°30\'45"', convert.ROUND_SECOND) == ('-', 12, 30, 45)
    assert convert.string_to_dms('-12°30\'45"', convert.ROUND_SECOND, True) == ('-', 12, 30, 45)


def test_string_to_dms_lat_rounded_degree():
    assert convert.string_to_dms('12N30.75', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('12N30.75', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.string_to_dms('12N30', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('12N30', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.string_to_dms('12S30.75', convert.ROUND_DEGREE) == ('-', 13)
    assert convert.string_to_dms('12S30.75', convert.ROUND_DEGREE, True) == ('-', 13, 0, 0)
    assert convert.string_to_dms('12S30', convert.ROUND_DEGREE) == ('-', 13)
    assert convert.string_to_dms('12S30', convert.ROUND_DEGREE, True) == ('-', 13, 0, 0)


def test_string_to_dms_lat_rounded_minute():
    assert convert.string_to_dms('12N30.75', convert.ROUND_MINUTE) == ('+', 12, 31)
    assert convert.string_to_dms('12N30.75', convert.ROUND_MINUTE, True) == ('+', 12, 31, 0)
    assert convert.string_to_dms('12N30', convert.ROUND_MINUTE) == ('+', 12, 30)
    assert convert.string_to_dms('12N30', convert.ROUND_MINUTE, True) == ('+', 12, 30, 0)
    assert convert.string_to_dms('12S30.75', convert.ROUND_MINUTE) == ('-', 12, 31)
    assert convert.string_to_dms('12S30.75', convert.ROUND_MINUTE, True) == ('-', 12, 31, 0)
    assert convert.string_to_dms('12S30', convert.ROUND_MINUTE) == ('-', 12, 30)
    assert convert.string_to_dms('12S30', convert.ROUND_MINUTE, True) == ('-', 12, 30, 0)


def test_string_to_dms_lat_rounded_second():
    assert convert.string_to_dms('12N30.75', convert.ROUND_SECOND) == ('+', 12, 30, 45)
    assert convert.string_to_dms('12N30.75', convert.ROUND_SECOND, True) == ('+', 12, 30, 45)
    assert convert.string_to_dms('12S30.75', convert.ROUND_SECOND) == ('-', 12, 30, 45)
    assert convert.string_to_dms('12S30.75', convert.ROUND_SECOND, True) == ('-', 12, 30, 45)


def test_string_to_dms_lon_rounded_degree():
    assert convert.string_to_dms('12E30.75', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('12E30.75', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.string_to_dms('12E30', convert.ROUND_DEGREE) == ('+', 13)
    assert convert.string_to_dms('12E30', convert.ROUND_DEGREE, True) == ('+', 13, 0, 0)
    assert convert.string_to_dms('12W30.75', convert.ROUND_DEGREE) == ('-', 13)
    assert convert.string_to_dms('12W30.75', convert.ROUND_DEGREE, True) == ('-', 13, 0, 0)
    assert convert.string_to_dms('12W30', convert.ROUND_DEGREE) == ('-', 13)
    assert convert.string_to_dms('12W30', convert.ROUND_DEGREE, True) == ('-', 13, 0, 0)


def test_string_to_dms_lon_rounded_minute():
    assert convert.string_to_dms('12E30.75', convert.ROUND_MINUTE) == ('+', 12, 31)
    assert convert.string_to_dms('12E30.75', convert.ROUND_MINUTE, True) == ('+', 12, 31, 0)
    assert convert.string_to_dms('12E30', convert.ROUND_MINUTE) == ('+', 12, 30)
    assert convert.string_to_dms('12E30', convert.ROUND_MINUTE, True) == ('+', 12, 30, 0)
    assert convert.string_to_dms('12W30.75', convert.ROUND_MINUTE) == ('-', 12, 31)
    assert convert.string_to_dms('12W30.75', convert.ROUND_MINUTE, True) == ('-', 12, 31, 0)
    assert convert.string_to_dms('12W30', convert.ROUND_MINUTE) == ('-', 12, 30)
    assert convert.string_to_dms('12W30', convert.ROUND_MINUTE, True) == ('-', 12, 30, 0)


def test_string_to_dms_lon_rounded_second():
    assert convert.string_to_dms('12E30.75', convert.ROUND_SECOND) == ('+', 12, 30, 45)
    assert convert.string_to_dms('12E30.75', convert.ROUND_SECOND, True) == ('+', 12, 30, 45)
    assert convert.string_to_dms('12W30.75', convert.ROUND_SECOND) == ('-', 12, 30, 45)
    assert convert.string_to_dms('12W30.75', convert.ROUND_SECOND, True) == ('-', 12, 30, 45)


""" dec_to_string() """
def test_dec_to_string_time_round_degree():
    assert convert.dec_to_string(12.5125, convert.FORMAT_TIME, convert.ROUND_DEGREE) == '13:00:00'
    assert convert.dec_to_string(12.5, convert.FORMAT_TIME, convert.ROUND_DEGREE) == '13:00:00'
    assert convert.dec_to_string(12, convert.FORMAT_TIME, convert.ROUND_DEGREE) == '12:00:00'


def test_dec_to_string_time_round_minute():
    assert convert.dec_to_string(12.5125, convert.FORMAT_TIME, convert.ROUND_MINUTE) == '12:31:00'
    assert convert.dec_to_string(12.5, convert.FORMAT_TIME, convert.ROUND_MINUTE) == '12:30:00'
    assert convert.dec_to_string(12, convert.FORMAT_TIME, convert.ROUND_MINUTE) == '12:00:00'


def test_dec_to_string_time_round_second():
    assert convert.dec_to_string(12.5125, convert.FORMAT_TIME, convert.ROUND_SECOND) == '12:30:45'
    assert convert.dec_to_string(12.5, convert.FORMAT_TIME, convert.ROUND_SECOND) == '12:30:00'
    assert convert.dec_to_string(12, convert.FORMAT_TIME, convert.ROUND_SECOND) == '12:00:00'


def test_dec_to_string_time_offset_round_degree():
    assert convert.dec_to_string(12.5125, convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '+13:00:00'
    assert convert.dec_to_string(12.5, convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '+13:00:00'
    assert convert.dec_to_string(12, convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '+12:00:00'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '-13:00:00'
    assert convert.dec_to_string(-12.5, convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '-13:00:00'
    assert convert.dec_to_string(-12, convert.FORMAT_TIME_OFFSET, convert.ROUND_DEGREE) == '-12:00:00'


def test_dec_to_string_time_offset_round_minute():
    assert convert.dec_to_string(12.5125, convert.FORMAT_TIME_OFFSET, convert.ROUND_MINUTE) == '+12:31:00'
    assert convert.dec_to_string(12.5, convert.FORMAT_TIME_OFFSET, convert.ROUND_MINUTE) == '+12:30:00'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_TIME_OFFSET, convert.ROUND_MINUTE) == '-12:31:00'
    assert convert.dec_to_string(-12.5, convert.FORMAT_TIME_OFFSET, convert.ROUND_MINUTE) == '-12:30:00'


def test_dec_to_string_time_offset_round_second():
    assert convert.dec_to_string(12.5125, convert.FORMAT_TIME_OFFSET, convert.ROUND_SECOND) == '+12:30:45'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_TIME_OFFSET, convert.ROUND_SECOND) == '-12:30:45'


def test_dec_to_string_dms_round_degree():
    assert convert.dec_to_string(12.5125, convert.FORMAT_DMS, convert.ROUND_DEGREE) == '13°'
    assert convert.dec_to_string(12.5, convert.FORMAT_DMS, convert.ROUND_DEGREE) == '13°'
    assert convert.dec_to_string(12, convert.FORMAT_DMS, convert.ROUND_DEGREE) == '12°'
    assert convert.dec_to_string(12.5125, convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '13°00\'00"'
    assert convert.dec_to_string(12.5, convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '13°00\'00"'
    assert convert.dec_to_string(12, convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '12°00\'00"'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_DMS, convert.ROUND_DEGREE) == '-13°'
    assert convert.dec_to_string(-12.5, convert.FORMAT_DMS, convert.ROUND_DEGREE) == '-13°'
    assert convert.dec_to_string(-12, convert.FORMAT_DMS, convert.ROUND_DEGREE) == '-12°'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '-13°00\'00"'
    assert convert.dec_to_string(-12.5, convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '-13°00\'00"'
    assert convert.dec_to_string(-12, convert.FORMAT_DMS, convert.ROUND_DEGREE, True) == '-12°00\'00"'


def test_dec_to_string_dms_round_minute():
    assert convert.dec_to_string(12.5125, convert.FORMAT_DMS, convert.ROUND_MINUTE) == '12°31\''
    assert convert.dec_to_string(12.5, convert.FORMAT_DMS, convert.ROUND_MINUTE) == '12°30\''
    assert convert.dec_to_string(12.5125, convert.FORMAT_DMS, convert.ROUND_MINUTE, True) == '12°31\'00"'
    assert convert.dec_to_string(12.5, convert.FORMAT_DMS, convert.ROUND_MINUTE, True) == '12°30\'00"'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_DMS, convert.ROUND_MINUTE) == '-12°31\''
    assert convert.dec_to_string(-12.5, convert.FORMAT_DMS, convert.ROUND_MINUTE) == '-12°30\''
    assert convert.dec_to_string(-12.5125, convert.FORMAT_DMS, convert.ROUND_MINUTE, True) == '-12°31\'00"'
    assert convert.dec_to_string(-12.5, convert.FORMAT_DMS, convert.ROUND_MINUTE, True) == '-12°30\'00"'


def test_dec_to_string_dms_round_second():
    assert convert.dec_to_string(12.5125, convert.FORMAT_DMS, convert.ROUND_SECOND) == '12°30\'45"'
    assert convert.dec_to_string(12.5, convert.FORMAT_DMS, convert.ROUND_SECOND) == '12°30\'00"'
    assert convert.dec_to_string(12.5125, convert.FORMAT_DMS, convert.ROUND_SECOND, True) == '12°30\'45"'
    assert convert.dec_to_string(12.5, convert.FORMAT_DMS, convert.ROUND_SECOND, True) == '12°30\'00"'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_DMS, convert.ROUND_SECOND) == '-12°30\'45"'
    assert convert.dec_to_string(-12.5, convert.FORMAT_DMS, convert.ROUND_SECOND) == '-12°30\'00"'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_DMS, convert.ROUND_SECOND, True) == '-12°30\'45"'
    assert convert.dec_to_string(-12.5, convert.FORMAT_DMS, convert.ROUND_SECOND, True) == '-12°30\'00"'


def test_dec_to_string_lat_round_degree():
    assert convert.dec_to_string(12.5125, convert.FORMAT_LAT, convert.ROUND_DEGREE) == '13N0.0'
    assert convert.dec_to_string(12.5, convert.FORMAT_LAT, convert.ROUND_DEGREE) == '13N0.0'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_LAT, convert.ROUND_DEGREE) == '13S0.0'
    assert convert.dec_to_string(-12.5, convert.FORMAT_LAT, convert.ROUND_DEGREE) == '13S0.0'


def test_dec_to_string_lat_round_minute():
    assert convert.dec_to_string(12.5125, convert.FORMAT_LAT, convert.ROUND_MINUTE) == '12N31.0'
    assert convert.dec_to_string(12.5, convert.FORMAT_LAT, convert.ROUND_MINUTE) == '12N30.0'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_LAT, convert.ROUND_MINUTE) == '12S31.0'
    assert convert.dec_to_string(-12.5, convert.FORMAT_LAT, convert.ROUND_MINUTE) == '12S30.0'


def test_dec_to_string_lat_round_second():
    assert convert.dec_to_string(12.5125, convert.FORMAT_LAT, convert.ROUND_SECOND) == '12N30.75'
    assert convert.dec_to_string(12.5, convert.FORMAT_LAT, convert.ROUND_SECOND) == '12N30.0'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_LAT, convert.ROUND_SECOND) == '12S30.75'
    assert convert.dec_to_string(-12.5, convert.FORMAT_LAT, convert.ROUND_SECOND) == '12S30.0'


def test_dec_to_string_lon_round_degree():
    assert convert.dec_to_string(12.5125, convert.FORMAT_LON, convert.ROUND_DEGREE) == '13E0.0'
    assert convert.dec_to_string(12.5, convert.FORMAT_LON, convert.ROUND_DEGREE) == '13E0.0'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_LON, convert.ROUND_DEGREE) == '13W0.0'
    assert convert.dec_to_string(-12.5, convert.FORMAT_LON, convert.ROUND_DEGREE) == '13W0.0'


def test_dec_to_string_lon_round_minute():
    assert convert.dec_to_string(12.5125, convert.FORMAT_LON, convert.ROUND_MINUTE) == '12E31.0'
    assert convert.dec_to_string(12.5, convert.FORMAT_LON, convert.ROUND_MINUTE) == '12E30.0'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_LON, convert.ROUND_MINUTE) == '12W31.0'
    assert convert.dec_to_string(-12.5, convert.FORMAT_LON, convert.ROUND_MINUTE) == '12W30.0'


def test_dec_to_string_lon_round_second():
    assert convert.dec_to_string(12.5125, convert.FORMAT_LON, convert.ROUND_SECOND) == '12E30.75'
    assert convert.dec_to_string(12.5, convert.FORMAT_LON, convert.ROUND_SECOND) == '12E30.0'
    assert convert.dec_to_string(-12.5125, convert.FORMAT_LON, convert.ROUND_SECOND) == '12W30.75'
    assert convert.dec_to_string(-12.5, convert.FORMAT_LON, convert.ROUND_SECOND) == '12W30.0'


""" string_to_dec() """
def test_string_to_dec_time():
    assert convert.string_to_dec('12:30:45') == 12.5125


def test_string_to_dec_time_offset():
    assert convert.string_to_dec('+12:30:45') == 12.5125
    assert convert.string_to_dec('-12:30:45') == -12.5125


def test_string_to_dec_dms():
    assert convert.string_to_dec('12°30\'45"') == 12.5125
    assert convert.string_to_dec('-12°30\'45"') == -12.5125


def test_string_to_dec_lat():
    assert convert.string_to_dec('12N30.75') == 12.5125
    assert convert.string_to_dec('12S30.75') == -12.5125


def test_string_to_dec_lon():
    assert convert.string_to_dec('12E30.75') == 12.5125
    assert convert.string_to_dec('12W30.75') == -12.5125


""" To save another few hundred tests the to_() functions are tested
to ensure they parse the input correctly and the accuracy of their
outputs is assumed to have already been covered by the above tests. """

def test_to_dec():
    assert convert.to_dec(12.5125) == 12.5125
    assert convert.to_dec(('+', 12, 30, 45)) == 12.5125
    assert convert.to_dec(['+', 12, 30, 45]) == 12.5125
    assert convert.to_dec('12.5125') == 12.5125
    assert convert.to_dec('12E30.75') == 12.5125

def test_to_dms():
    assert convert.to_dms(12.5125) == ('+', 12, 30, 45)
    assert convert.to_dms(('+', 12, 30, 45)) == ('+', 12, 30, 45)
    assert convert.to_dms(['+', 12, 30, 45]) == ('+', 12, 30, 45)
    assert convert.to_dms('12.5125') == ('+', 12, 30, 45)
    assert convert.to_dms('12E30.75') == ('+', 12, 30, 45)
    # Quick tests to ensure it respects args
    assert convert.to_dms(12.5125, convert.ROUND_MINUTE) == ('+', 12, 31)
    assert convert.to_dms(12.5125, convert.ROUND_MINUTE, True) == ('+', 12, 31, 0)

def test_to_string():
    assert convert.to_string(12.5125) == '12°30\'45"'
    assert convert.to_string(('+', 12, 30, 45)) == '12°30\'45"'
    assert convert.to_string(['+', 12, 30, 45]) == '12°30\'45"'
    assert convert.to_string('12.5125') == '12°30\'45"'
    assert convert.to_string('12E30.75') == '12°30\'45"'
    # Quick tests to ensure it respects args
    assert convert.to_string(12.5125, convert.FORMAT_LAT) == '12N30.75'
    assert convert.to_string(12.5125, convert.FORMAT_DMS, convert.ROUND_MINUTE) == '12°31\''
    assert convert.to_string(12.5125, convert.FORMAT_DMS, convert.ROUND_MINUTE, True) == '12°31\'00"'



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    This module contains miscellaneous calculations to ascertain basic chart
    and object information from the data returned from the ephemeris module.
    These functions will accept both an object and a float.

"""

import swisseph as swe

from immanuel.const import calc
from immanuel.tools import ephemeris


def moon_phase(sun: dict | float, moon: dict | float) -> int:
    """ Returns the moon phase given the positions of the Sun and Moon. """
    sun_lon, moon_lon = (object['lon'] if isinstance(object, dict) else object for object in (sun, moon))
    distance = swe.difdegn(moon_lon, sun_lon)

    for angle in range(45, 361, 45):
        if distance < angle:
            return angle


def is_daytime(sun: dict | float, asc: dict | float) -> bool:
    """ Returns whether the sun is above the ascendant. """
    sun_lon, asc_lon = (object['lon'] if isinstance(object, dict) else object for object in (sun, asc))
    return swe.difdeg2n(sun_lon, asc_lon) < 0


def pars_fortuna_longitude(sun: dict | float, moon: dict | float, asc: dict | float, formula: int) -> float:
    """ Returns the Part of Fortune longitude. """
    sun_lon, moon_lon, asc_lon = (object['lon'] if isinstance(object, dict) else object for object in (sun, moon, asc))

    if formula == calc.NIGHT_FORMULA or (formula == calc.DAY_NIGHT_FORMULA and not is_daytime(sun_lon, asc_lon)):
        lon = asc_lon + sun_lon - moon_lon
    else:
        lon = asc_lon + moon_lon - sun_lon

    return swe.degnorm(lon)


def sidereal_time(armc: dict | float) -> float:
    """ Returns sidereal time based on ARMC longitude. """
    return (armc['lon'] if isinstance(armc, dict) else armc) / 15


def object_movement(object: dict | float) -> int:
    """ Returns whether a chart object is direct, stationary or retrograde. """
    speed = object['speed'] if isinstance(object, dict) else object

    if -calc.STATION_SPEED <= speed <= calc.STATION_SPEED:
        return calc.STATIONARY

    return calc.DIRECT if speed > calc.STATION_SPEED else calc.RETROGRADE


def is_out_of_bounds(object: dict | float, jd: float = None, obliquity: float = None) -> bool:
    """ Returns whether the passed object is out of bounds either on the passed
    Julian date or relative to the passed obliquity. """
    if isinstance(object, dict):
        if 'dec' not in object:
            return None
        dec = object['dec']
    else:
        dec = object

    if jd is not None:
        obliquity = ephemeris.obliquity(jd)
    elif obliquity is None:
        return None

    return not -obliquity < dec < obliquity


def solar_year_length(jd: float) -> float:
    """ Returns the length in days of the year passed in the given
    Julian date. This is a direct copy of astro.com's calculations. """
    t = (jd - calc.J2000) / 365250
    t2 = t * t
    t3 = t2 * t
    t4 = t3 * t
    t5 = t4 * t
    # Arcsec per milllennium
    dvel = 1296027711.03429 + 2 * 109.15809 * t + 3 * 0.07207 * t2 - 4 * 0.23530 * t3 - 5 * 0.00180 * t4 + 6 * 0.00020 * t5
    # Degrees per millennium
    dvel /= 3600
    return 360 * 365250 / dvel



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Defines essential dignities and their associated tables.

"""

from immanuel.const import chart


""" Indices for reference. """
RULER = 0
EXALTED = 1
TRIPLICITY_RULER = 2
TERM_RULER = 3
FACE_RULER = 4
MUTUAL_RECEPTION_RULER = 5
MUTUAL_RECEPTION_EXALTED = 6
MUTUAL_RECEPTION_TRIPLICITY_RULER = 7
MUTUAL_RECEPTION_TERM_RULER = 8
MUTUAL_RECEPTION_FACE_RULER = 9
IN_RULERSHIP_ELEMENT = 10
DETRIMENT = 11
FALL = 12
PEREGRINE = 13


""" Modern rulerships. """
MODERN_RULERSHIPS = {
    chart.ARIES: chart.MARS,
    chart.TAURUS: chart.VENUS,
    chart.GEMINI: chart.MERCURY,
    chart.CANCER: chart.MOON,
    chart.LEO: chart.SUN,
    chart.VIRGO: chart.MERCURY,
    chart.LIBRA: chart.VENUS,
    chart.SCORPIO: chart.PLUTO,
    chart.SAGITTARIUS: chart.JUPITER,
    chart.CAPRICORN: chart.SATURN,
    chart.AQUARIUS: chart.URANUS,
    chart.PISCES: chart.NEPTUNE,
}


""" Traditional rulerships. """
TRADITIONAL_RULERSHIPS = {
    chart.ARIES: chart.MARS,
    chart.TAURUS: chart.VENUS,
    chart.GEMINI: chart.MERCURY,
    chart.CANCER: chart.MOON,
    chart.LEO: chart.SUN,
    chart.VIRGO: chart.MERCURY,
    chart.LIBRA: chart.VENUS,
    chart.SCORPIO: chart.MARS,
    chart.SAGITTARIUS: chart.JUPITER,
    chart.CAPRICORN: chart.SATURN,
    chart.AQUARIUS: chart.SATURN,
    chart.PISCES: chart.JUPITER,
}


""" Planets in exaltation. """
EXALTATIONS = {
    chart.ARIES: chart.SUN,
    chart.TAURUS: chart.MOON,
    chart.GEMINI: None,
    chart.CANCER: chart.JUPITER,
    chart.LEO: chart.NEPTUNE,
    chart.VIRGO: chart.MERCURY,
    chart.LIBRA: chart.SATURN,
    chart.SCORPIO: chart.URANUS,
    chart.SAGITTARIUS: None,
    chart.CAPRICORN: chart.MARS,
    chart.AQUARIUS: None,
    chart.PISCES: chart.VENUS,
}


""" Face rulers. """
FACE_RULERS = {
    chart.ARIES: (chart.MARS, chart.SUN, chart.VENUS),
    chart.TAURUS: (chart.MERCURY, chart.MOON, chart.SATURN),
    chart.GEMINI: (chart.JUPITER, chart.MARS, chart.SUN),
    chart.CANCER: (chart.VENUS, chart.MERCURY, chart.MOON),
    chart.LEO: (chart.SATURN, chart.JUPITER, chart.MARS),
    chart.VIRGO: (chart.SUN, chart.VENUS, chart.MERCURY),
    chart.LIBRA: (chart.MOON, chart.SATURN, chart.JUPITER),
    chart.SCORPIO: (chart.MARS, chart.SUN, chart.VENUS),
    chart.SAGITTARIUS: (chart.MERCURY, chart.MOON, chart.SATURN),
    chart.CAPRICORN: (chart.JUPITER, chart.MARS, chart.SUN),
    chart.AQUARIUS: (chart.VENUS, chart.MERCURY, chart.MOON),
    chart.PISCES: (chart.SATURN, chart.JUPITER, chart.MARS),
}


""" Ptolemy's terms. """
PTOLEMAIC_TERMS = {
    chart.ARIES: {
        chart.JUPITER: (0, 6),
        chart.VENUS: (6, 14),
        chart.MERCURY: (14, 21),
        chart.MARS: (21, 26),
        chart.SATURN: (26, 30),
    },
    chart.TAURUS: {
        chart.VENUS: (0, 8),
        chart.MERCURY: (8, 15),
        chart.JUPITER: (15, 22),
        chart.SATURN: (22, 26),
        chart.MARS: (26, 30),
    },
    chart.GEMINI: {
        chart.MERCURY: (0, 7),
        chart.JUPITER: (7, 14),
        chart.VENUS: (14, 21),
        chart.SATURN: (21, 25),
        chart.MARS: (25, 30),
    },
    chart.CANCER: {
        chart.MARS: (0, 6),
        chart.JUPITER: (6, 13),
        chart.MERCURY: (13, 20),
        chart.VENUS: (20, 27),
        chart.SATURN: (27, 30),
    },
    chart.LEO: {
        chart.SATURN: (0, 6),
        chart.MERCURY: (6, 13),
        chart.VENUS: (13, 19),
        chart.JUPITER: (19, 25),
        chart.MARS: (25, 30),
    },
    chart.VIRGO: {
        chart.MERCURY: (0, 7),
        chart.VENUS: (7, 13),
        chart.JUPITER: (13, 18),
        chart.SATURN: (18, 24),
        chart.MARS: (24, 30),
    },
    chart.LIBRA: {
        chart.SATURN: (0, 6),
        chart.VENUS: (6, 11),
        chart.JUPITER: (11, 19),
        chart.MERCURY: (19, 24),
        chart.MARS: (24, 30),
    },
    chart.SCORPIO: {
        chart.MARS: (0, 6),
        chart.JUPITER: (6, 14),
        chart.VENUS: (14, 21),
        chart.MERCURY: (21, 27),
        chart.SATURN: (27, 30),
    },
    chart.SAGITTARIUS: {
        chart.JUPITER: (0, 8),
        chart.VENUS: (8, 14),
        chart.MERCURY: (14, 19),
        chart.SATURN: (19, 25),
        chart.MARS: (25, 30),
    },
    chart.CAPRICORN: {
        chart.VENUS: (0, 6),
        chart.MERCURY: (6, 12),
        chart.JUPITER: (12, 19),
        chart.MARS: (19, 25),
        chart.SATURN: (25, 30),
    },
    chart.AQUARIUS: {
        chart.SATURN: (0, 6),
        chart.MERCURY: (6, 12),
        chart.VENUS: (12, 20),
        chart.JUPITER: (20, 25),
        chart.MARS: (25, 30),
    },
    chart.PISCES: {
        chart.VENUS: (0, 8),
        chart.JUPITER: (8, 14),
        chart.MERCURY: (14, 20),
        chart.MARS: (20, 26),
        chart.SATURN: (26, 30),
    },
}


""" Egyptian terms. """
EGYPTIAN_TERMS = {
    chart.ARIES: {
        chart.JUPITER: (0, 6),
        chart.VENUS: (6, 12),
        chart.MERCURY: (12, 20),
        chart.MARS: (20, 25),
        chart.SATURN: (25, 30),
    },
    chart.TAURUS: {
        chart.VENUS: (0, 8),
        chart.MERCURY: (8, 14),
        chart.JUPITER: (14, 22),
        chart.SATURN: (22, 27),
        chart.MARS: (27, 30),
    },
    chart.GEMINI: {
        chart.MERCURY: (0, 6),
        chart.JUPITER: (6, 12),
        chart.VENUS: (12, 17),
        chart.MARS: (17, 24),
        chart.SATURN: (24, 30),
    },
    chart.CANCER: {
        chart.MARS: (0, 7),
        chart.VENUS: (7, 13),
        chart.MERCURY: (13, 19),
        chart.JUPITER: (19, 26),
        chart.SATURN: (26, 30),
    },
    chart.LEO: {
        chart.JUPITER: (0, 6),
        chart.VENUS: (6, 11),
        chart.SATURN: (11, 18),
        chart.MERCURY: (18, 24),
        chart.MARS: (24, 30),
    },
    chart.VIRGO: {
        chart.MERCURY: (0, 7),
        chart.VENUS: (7, 17),
        chart.JUPITER: (17, 21),
        chart.MARS: (21, 28),
        chart.SATURN: (28, 30),
    },
    chart.LIBRA: {
        chart.SATURN: (0, 6),
        chart.MERCURY: (6, 14),
        chart.JUPITER: (14, 21),
        chart.VENUS: (21, 28),
        chart.MARS: (28, 30),
    },
    chart.SCORPIO: {
        chart.MARS: (0, 7),
        chart.VENUS: (7, 11),
        chart.MERCURY: (11, 19),
        chart.JUPITER: (19, 24),
        chart.SATURN: (24, 30),
    },
    chart.SAGITTARIUS: {
        chart.JUPITER: (0, 12),
        chart.VENUS: (12, 17),
        chart.MERCURY: (17, 21),
        chart.SATURN: (21, 26),
        chart.MARS: (26, 30),
    },
    chart.CAPRICORN: {
        chart.MERCURY: (0, 7),
        chart.JUPITER: (7, 14),
        chart.VENUS: (14, 22),
        chart.SATURN: (22, 26),
        chart.MARS: (26, 30),
    },
    chart.AQUARIUS: {
        chart.MERCURY: (0, 7),
        chart.VENUS: (7, 13),
        chart.JUPITER: (13, 20),
        chart.MARS: (20, 25),
        chart.SATURN: (25, 30),
    },
    chart.PISCES: {
        chart.VENUS: (0, 12),
        chart.JUPITER: (12, 16),
        chart.MERCURY: (16, 19),
        chart.MARS: (19, 28),
        chart.SATURN: (28, 30),
    },
}


""" Ptolemy's triplicities. """
PTOLEMAIC_TRIPLICITIES = {
    chart.ARIES: {
        'day': chart.SUN,
        'night': chart.JUPITER,
        'participatory': chart.MARS,
    },
    chart.TAURUS: {
        'day': chart.VENUS,
        'night': chart.MOON,
        'participatory': chart.SATURN,
    },
    chart.GEMINI: {
        'day': chart.SATURN,
        'night': chart.MERCURY,
        'participatory': chart.JUPITER,
    },
    chart.CANCER: {
        'day': chart.VENUS,
        'night': chart.MOON,
        'participatory': chart.MARS,
    },
    chart.LEO: {
        'day': chart.SUN,
        'night': chart.JUPITER,
        'participatory': chart.MARS,
    },
    chart.VIRGO: {
        'day': chart.VENUS,
        'night': chart.MOON,
        'participatory': chart.SATURN,
    },
    chart.LIBRA: {
        'day': chart.SATURN,
        'night': chart.MERCURY,
        'participatory': chart.JUPITER,
    },
    chart.SCORPIO: {
        'day': chart.VENUS,
        'night': chart.MOON,
        'participatory': chart.MARS,
    },
    chart.SAGITTARIUS: {
        'day': chart.SUN,
        'night': chart.JUPITER,
        'participatory': chart.MARS,
    },
    chart.CAPRICORN: {
        'day': chart.VENUS,
        'night': chart.MOON,
        'participatory': chart.SATURN,
    },
    chart.AQUARIUS: {
        'day': chart.SATURN,
        'night': chart.MERCURY,
        'participatory': chart.JUPITER,
    },
    chart.PISCES: {
        'day': chart.VENUS,
        'night': chart.MOON,
        'participatory': chart.MARS,
    },
}


""" Lilly's triplicities. """
LILLEAN_TRIPLICITIES = {
    chart.ARIES: {
        'day': chart.SUN,
        'night': chart.JUPITER,
    },
    chart.TAURUS: {
        'day': chart.VENUS,
        'night': chart.MOON,
    },
    chart.GEMINI: {
        'day': chart.SATURN,
        'night': chart.MERCURY,
    },
    chart.CANCER: {
        'day': chart.MARS,
        'night': chart.MARS,
    },
    chart.LEO: {
        'day': chart.SUN,
        'night': chart.JUPITER,
    },
    chart.VIRGO: {
        'day': chart.VENUS,
        'night': chart.MOON,
    },
    chart.LIBRA: {
        'day': chart.SATURN,
        'night': chart.MERCURY,
    },
    chart.SCORPIO: {
        'day': chart.MARS,
        'night': chart.MARS,
    },
    chart.SAGITTARIUS: {
        'day': chart.SUN,
        'night': chart.JUPITER,
    },
    chart.CAPRICORN: {
        'day': chart.VENUS,
        'night': chart.MOON,
    },
    chart.AQUARIUS: {
        'day': chart.SATURN,
        'night': chart.MERCURY,
    },
    chart.PISCES: {
        'day': chart.MARS,
        'night': chart.MARS,
    },
}


""" Dorotheus' triplicities. """
DOROTHEAN_TRIPLICITIES = {
    chart.ARIES: {
        'day': chart.SUN,
        'night': chart.JUPITER,
        'participatory': chart.SATURN,
    },
    chart.TAURUS: {
        'day': chart.VENUS,
        'night': chart.MOON,
        'participatory': chart.MARS,
    },
    chart.GEMINI: {
        'day': chart.SATURN,
        'night': chart.MERCURY,
        'participatory': chart.JUPITER,
    },
    chart.CANCER: {
        'day': chart.VENUS,
        'night': chart.MARS,
        'participatory': chart.MOON,
    },
    chart.LEO: {
        'day': chart.SUN,
        'night': chart.JUPITER,
        'participatory': chart.SATURN,
    },
    chart.VIRGO: {
        'day': chart.VENUS,
        'night': chart.MOON,
        'participatory': chart.MARS,
    },
    chart.LIBRA: {
        'day': chart.SATURN,
        'night': chart.MERCURY,
        'participatory': chart.JUPITER,
    },
    chart.SCORPIO: {
        'day': chart.VENUS,
        'night': chart.MARS,
        'participatory': chart.MOON,
    },
    chart.SAGITTARIUS: {
        'day': chart.SUN,
        'night': chart.JUPITER,
        'participatory': chart.SATURN,
    },
    chart.CAPRICORN: {
        'day': chart.VENUS,
        'night': chart.MOON,
        'participatory': chart.MARS,
    },
    chart.AQUARIUS: {
        'day': chart.SATURN,
        'night': chart.MERCURY,
        'participatory': chart.JUPITER,
    },
    chart.PISCES: {
        'day': chart.VENUS,
        'night': chart.MARS,
        'participatory': chart.MOON,
    },
}



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    The position module's calculations are tested against
    figures from astro.com. The sign_longitude() function will already
    have been partially tested in the ephemeris module's tests since
    there is no other way to convert the ephemeris module's longitudes
    into sign-specific longitudes.

"""

import os
from datetime import datetime

from pytest import fixture

from immanuel import setup
from immanuel.const import calc, chart
from immanuel.tools import convert, date, ephemeris, position


@fixture
def coords():
    # San Diego coords as used by astro.com
    return [convert.string_to_dec(v) for v in ('32n43', '117w09')]

@fixture
def jd(coords):
    return date.to_jd('2000-01-01 10:00', *coords)

@fixture
def data(jd, coords):
    setup.add_filepath(os.path.dirname(__file__))

    return {
        'asc': ephemeris.angle(chart.ASC, jd, *coords, chart.PLACIDUS),
        'house_2': ephemeris.house(chart.HOUSE2, jd, *coords, chart.PLACIDUS),
        'sun': ephemeris.planet(chart.SUN, jd),
        'pof': ephemeris.point(chart.PARS_FORTUNA, jd, *coords, chart.PLACIDUS, calc.DAY_NIGHT_FORMULA),
        'juno': ephemeris.asteroid(chart.JUNO, jd),       # Included with planets
        'lilith': ephemeris.asteroid(1181, jd),           # From its own file
        'antares': ephemeris.fixed_star('Antares', jd),
    }

@fixture
def astro():
    """ Results copied from astro.com chart table. """
    return {
        # angle
        'asc': {
            'sign' : chart.PISCES,
            'opposite_sign': chart.VIRGO,
            'lon': '05°36\'38"',
            'decan': chart.DECAN1,
            'element': chart.WATER,
            'modality': chart.MUTABLE,
        },
        # house
        'house_2': {
            'sign': chart.ARIES,
            'opposite_sign': chart.LIBRA,
            'lon': '17°59\'40"',
            'decan': chart.DECAN2,
            'element': chart.FIRE,
            'modality': chart.CARDINAL,
        },
        # planet
        'sun': {
            'sign': chart.CAPRICORN,
            'opposite_sign': chart.CANCER,
            'house': 11,
            'opposite_house': 5,
            'lon': '10°37\'26"',
            'decan': chart.DECAN2,
            'element': chart.EARTH,
            'modality': chart.CARDINAL,
        },
        # point
        'pof': {
            'sign': chart.CAPRICORN,
            'opposite_sign': chart.CANCER,
            'house': 11,
            'opposite_house': 5,
            'lon': '11°18\'41"',
            'decan': chart.DECAN2,
            'element': chart.EARTH,
            'modality': chart.CARDINAL,
        },
        # default asteroid
        'juno': {
            'sign': chart.CAPRICORN,
            'opposite_sign': chart.CANCER,
            'house': 11,
            'opposite_house': 5,
            'lon': '08°05\'21"',
            'decan': chart.DECAN1,
            'element': chart.EARTH,
            'modality': chart.CARDINAL,
        },
        # external asteroid
        'lilith': {
            'sign': chart.PISCES,
            'opposite_sign': chart.VIRGO,
            'house': 1,
            'opposite_house': 7,
            'lon': '18°16\'47"',
            'decan': chart.DECAN2,
            'element': chart.WATER,
            'modality': chart.MUTABLE,
        },
        # fixed star
        'antares': {
            'sign': chart.SAGITTARIUS,
            'opposite_sign': chart.GEMINI,
            'house': 9,
            'opposite_house': 3,
            'lon': '09°45\'12"',
            'decan': chart.DECAN1,
            'element': chart.FIRE,
            'modality': chart.MUTABLE,
        },
    }


def test_sign(data, astro):
    for key, object in data.items():
        assert position.sign(object) == astro[key]['sign']


def test_sign_longitude(data, astro):
    for key, object in data.items():
        lon = position.sign_longitude(object)
        assert convert.dec_to_string(lon) == astro[key]['lon']


def test_opposite_sign(data, astro):
    for key, object in data.items():
        assert position.opposite_sign(object) == astro[key]['opposite_sign']


def test_decan(data, astro):
    for key, object in data.items():
        assert position.decan(object) == astro[key]['decan']


def test_house(jd, coords, data, astro):
    houses = ephemeris.houses(jd, *coords, chart.PLACIDUS)

    for key, object in {k: v for k, v in data.items() if 'house' in v}:
        assert position.house(object, houses) == astro[key]['house']


def test_opposite_house(jd, coords, data, astro):
    houses = ephemeris.houses(jd, *coords, chart.PLACIDUS)

    for key, object in {k: v for k, v in data.items() if 'house' in v}:
        assert position.opposite_house(object, houses) == astro[key]['opposite_house']


def test_element(data, astro):
    for key, object in data.items():
        assert position.element(object) == astro[key]['element']


def test_modality(data, astro):
    for key, object in data.items():
        assert position.modality(object) == astro[key]['modality']



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Defines flexible classes to represent data in multiple formats.
    While user-friendly names are defined in the const.names module,
    JSON keys are defined here, either explicitly or as class members.

"""

from datetime import datetime

from immanuel.const import calc, chart, dignities, names
from immanuel.reports import dignity
from immanuel.setup import settings
from immanuel.tools import calculate, convert, date, ephemeris, position


class Angle:
    def __init__(self, angle: float, format: int = convert.FORMAT_DMS) -> None:
        self.raw = angle
        self.formatted = convert.dec_to_string(angle, format)
        self.direction = None
        self.degrees = None
        self.minutes = None
        self.seconds = None
        self.__dict__.update(dict(zip(('direction', 'degrees', 'minutes', 'seconds'), convert.dec_to_dms(self.raw))))

    def __str__(self) -> str:
        return self.formatted


class Aspect:
    def __init__(self, aspect: dict, objects: dict) -> None:
        self._active_name = objects[aspect['active']]['name']
        self._passive_name = objects[aspect['passive']]['name']
        self.active = aspect['active']
        self.passive = aspect['passive']
        self.type = names.ASPECTS[aspect['aspect']]
        self.aspect = aspect['aspect']
        self.orb = aspect['orb']
        self.distance = Angle(aspect['distance'])
        self.difference = Angle(aspect['difference'])
        self.movement = AspectMovement(aspect['movement'])
        self.condition = AspectCondition(aspect['condition'])

    def __str__(self) -> str:
        return f'{self._active_name} {self._passive_name} {self.type} within {self.difference} ({self.movement}, {self.condition})'


class AspectCondition:
    def __init__(self, condition: int) -> None:
        self.associate = condition == calc.ASSOCIATE
        self.dissociate = condition == calc.DISSOCIATE
        self.formatted = names.ASPECT_CONDITIONS[condition]

    def __str__(self) -> str:
        return self.formatted


class AspectMovement:
    def __init__(self, movement: int) -> None:
        self.applicative = movement == calc.APPLICATIVE
        self.exact = movement == calc.EXACT
        self.separative = movement == calc.SEPARATIVE
        self.formatted = names.ASPECT_MOVEMENTS[movement]

    def __str__(self) -> str:
        return self.formatted


class Coordinates:
    def __init__(self, latitude: float, longitude: float) -> None:
        self.latitude = Angle(latitude, convert.FORMAT_LAT)
        self.longitude = Angle(longitude, convert.FORMAT_LON)

    def __str__(self) -> str:
        return f'{self.latitude}, {self.longitude}'


class DateTime:
    def __init__(self, dt: datetime | float, armc: dict | float = None, latitude: float = None, longitude: float = None, time_is_dst: bool = None) -> None:
        self.datetime = date.to_datetime(dt, latitude, longitude)
        self.timezone = self.datetime.tzname()
        self.ambiguous = date.ambiguous(self.datetime) and time_is_dst is None
        self.julian = date.to_jd(dt)
        self.deltat = ephemeris.deltat(self.julian)

        if armc is not None:
            self.sidereal_time = convert.dec_to_string(calculate.sidereal_time(armc), convert.FORMAT_TIME)

    def __str__(self) -> str:
        str = f'{self.datetime.strftime("%a %b %d %Y %I:%M:%S %p")} {self.timezone}'

        if self.ambiguous:
            str += ' (ambiguous)'

        return str


class DignityState:
    def __init__(self, dignity_state: dict) -> None:
        self.ruler = dignity_state[dignities.RULER]
        self.exalted = dignity_state[dignities.EXALTED]
        self.triplicity_ruler = dignity_state[dignities.TRIPLICITY_RULER]
        self.term_ruler = dignity_state[dignities.TERM_RULER]
        self.face_ruler = dignity_state[dignities.FACE_RULER]
        self.mutual_reception_ruler = dignity_state[dignities.MUTUAL_RECEPTION_RULER]
        self.mutual_reception_exalted = dignity_state[dignities.MUTUAL_RECEPTION_EXALTED]
        self.mutual_reception_triplicity_ruler = dignity_state[dignities.MUTUAL_RECEPTION_TRIPLICITY_RULER]
        self.mutual_reception_term_ruler = dignity_state[dignities.MUTUAL_RECEPTION_TERM_RULER]
        self.mutual_reception_face_ruler = dignity_state[dignities.MUTUAL_RECEPTION_FACE_RULER]
        self.detriment = dignity_state[dignities.DETRIMENT]
        self.fall = dignity_state[dignities.FALL]
        self.peregrine = dignity_state[dignities.PEREGRINE]
        self.formatted = [names.DIGNITIES[dignity] for dignity, active in dignity_state.items() if active]

    def __str__(self) -> str:
        return ', '.join(self.formatted)


class EclipseType:
    def __init__(self, eclipse_type: int) -> None:
        self.total = eclipse_type == chart.TOTAL
        self.annular = eclipse_type == chart.ANNULAR
        self.partial = eclipse_type == chart.PARTIAL
        self.annular_total = eclipse_type == chart.ANNULAR_TOTAL
        self.penumbral = eclipse_type == chart.PENUMBRAL
        self.formatted = names.ECLIPSE_TYPES[eclipse_type]

    def __str__(self) -> str:
        return self.formatted


class House:
    def __init__(self, house: dict) -> None:
        self.index = house['index']
        self.number = house['number']
        self.name = house['name']

    def __str__(self) -> str:
        return self.name


class MoonPhase:
    def __init__(self, moon_phase: int) -> None:
        self.new_moon = moon_phase == calc.NEW_MOON
        self.waxing_crescent = moon_phase == calc.WAXING_CRESCENT
        self.first_quarter = moon_phase == calc.FIRST_QUARTER
        self.waxing_gibbous = moon_phase == calc.WAXING_GIBBOUS
        self.full_moon = moon_phase == calc.FULL_MOON
        self.disseminating = moon_phase == calc.DISSEMINATING
        self.third_quarter = moon_phase == calc.THIRD_QUARTER
        self.balsamic = moon_phase == calc.BALSAMIC
        self.formatted = names.MOON_PHASES[moon_phase]

    def __str__(self) -> str:
        return self.formatted


class Object:
    def __init__(self, object: dict, objects: dict = None, houses: dict = None, is_daytime: bool = None, obliquity: float = None) -> None:
        self.index = object['index']
        self.name = object['name']
        self.type = ObjectType(object['type'])

        if 'eclipse_type' in object:
            self.eclipse_type = EclipseType(object['eclipse_type'])

        if 'date_time' in object:
            self.date_time = DateTime(object['date_time'])

        if 'lat' in object:
            self.latitude = Angle(object['lat'])

        self.longitude = Angle(object['lon'])
        self.sign_longitude = Angle(position.sign_longitude(object))
        self.sign = Sign(position.sign(object))

        if houses is not None:
            self.house = House(position.house(object, houses))

        if 'dist' in object:
            self.distance = object['dist']

        self.speed = object['speed']

        if object['type'] not in (chart.HOUSE, chart.ANGLE, chart.FIXED_STAR):
            self.movement = ObjectMovement(object)

            if 'dec' in object:
                self.declination = Angle(object['dec'])
                self.out_of_bounds = calculate.is_out_of_bounds(object=object, obliquity=obliquity)

        if 'size' in object:
            self.size = object['size']

        if objects is not None and object['type'] == chart.PLANET and is_daytime is not None and calc.PLANETS.issubset(objects):
            dignity_state = dignity.all(object=object, objects=objects, is_daytime=is_daytime)
            self.dignities = DignityState(dignity_state)
            self.score = dignity.score(dignity_state)

    def __str__(self) -> str:
        str = f'{self.name} {self.sign_longitude} in {self.sign}'

        if hasattr(self, 'house'):
            str += f', {self.house.name}'

        return str


class ObjectMovement:
    def __init__(self, object: dict) -> None:
        self._movement = calculate.object_movement(object)
        self.direct = self._movement == calc.DIRECT
        self.stationary = self._movement == calc.STATIONARY
        self.retrograde = self._movement == calc.RETROGRADE
        self.formatted = names.OBJECT_MOVEMENTS[self._movement]

    def __str__(self) -> str:
        return self.formatted


class ObjectType:
    def __init__(self, type: int) -> None:
        self.index = type
        self.name = names.OBJECTS[type]

    def __str__(self) -> str:
        return self.name


class Sign:
    def __init__(self, number: int) -> None:
        self.number = number
        self.name = names.SIGNS[self.number]

    def __str__(self) -> str:
        return self.name


class Subject:
    def __init__(self, subject: 'Subject') -> None:
        armc = ephemeris.angle(
                index=chart.ARMC,
                jd=subject.julian_date,
                lat=subject.latitude,
                lon=subject.longitude,
                house_system=settings.house_system,
            )
        self.date_time = DateTime(
                dt=subject.date_time,
                armc=armc,
                latitude=subject.latitude,
                longitude=subject.longitude,
                time_is_dst=subject.time_is_dst,
            )
        self.coordinates = Coordinates(
                latitude=subject.latitude,
                longitude=subject.longitude,
            )

    def __str__(self) -> str:
        return f'{self.date_time} at {self.coordinates}'


class Elements:
    def __init__(self, elements: dict) -> None:
        self.fire = elements[chart.FIRE]
        self.earth = elements[chart.EARTH]
        self.air = elements[chart.AIR]
        self.water = elements[chart.WATER]

    def __str__(self) -> str:
        return f'Fire: {len(self.fire)}, Earth: {len(self.earth)}, Air: {len(self.air)}, Water: {len(self.water)}'


class Modalities:
    def __init__(self, modalities: dict) -> None:
        self.cardinal = modalities[chart.CARDINAL]
        self.fixed = modalities[chart.FIXED]
        self.mutable = modalities[chart.MUTABLE]

    def __str__(self) -> str:
        return f'Cardinal: {len(self.cardinal)}, Fixed: {len(self.fixed)}, Mutable: {len(self.mutable)}'


class Quadrants:
    def __init__(self, quadrants: dict) -> None:
        self.first = quadrants[1]
        self.second = quadrants[2]
        self.third = quadrants[3]
        self.fourth = quadrants[4]

    def __str__(self) -> str:
        return f'First: {len(self.first)}, Second: {len(self.second)}, Third: {len(self.third)}, Fourth: {len(self.fourth)}'



"""
    This file is part of immanuel - (C) The Rift Lab
    Author: Robert Davies (robert@theriftlab.com)


    Extracts chart-shape and aspect patterns from a dict of chart objects
    provided by the ephemeris module.

"""

import swisseph as swe

from immanuel.const import calc
from immanuel.setup import settings


def chart_shape(objects: dict) -> int:
    # Filter objects
    objects = { k: v for k, v in objects.items() if k in settings.chart_shape_objects }

    # Sort objects by longitude
    longitudes = sorted([v['lon'] for v in objects.values()])
    diffs = [swe.difdegn(_next(longitudes, k), v) for k, v in enumerate(longitudes)]
    max_diff = max(diffs)

    # All planets within 120º can only be a bundle
    if max_diff >= 240-settings.chart_shape_orb:
        return calc.BUNDLE

    # Bucket handle planet(s) must be at least 90º from edges of main cluster
    for k, v in enumerate(diffs):
        next = _next(diffs, k)
        second_next = _next(diffs, k, 2)

        if v >= 90-settings.chart_shape_orb and (next >= 90-settings.chart_shape_orb or (next <= settings.chart_shape_orb and second_next >= 90-settings.chart_shape_orb)):
            return calc.BUCKET

    # All planets being within 180º with no bucket handle means a bowl
    if max_diff >= 180-settings.chart_shape_orb:
        return calc.BOWL

    # All planets being within 240º with no bucket handle means a locomotive
    if max_diff >= 120-settings.chart_shape_orb:
        return calc.LOCOMOTIVE

    diffs.sort()

    # Only two gaps of at least 60º mean a seesaw
    if len([v for v in diffs if v >= 60-settings.chart_shape_orb]) == 2:
        return calc.SEESAW

    # Three gaps of at least 30º mean a splay
    if len([v for v in diffs if v >= 30-settings.chart_shape_orb]) == 3:
        return calc.SPLAY

    # Default to no particular pattern
    return calc.SPLASH


def _next(data: list, key: int, step: int = 1) -> float:
    return data[(key+step)%(len(data))]
